package org.jboss.seam.selectitems.jsf;

import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;

import javax.faces.component.StateHolder;
import javax.faces.component.UIComponent;
import javax.faces.context.FacesContext;
import javax.faces.convert.ConverterException;
import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;

import org.jboss.seam.selectitems.SelectItemsConfig;
import org.jboss.seam.util.Naming;

public abstract class EntityConverter<E> extends SelectItemsConverter implements
		StateHolder {
	
	public static final String EM_ACCESS_ERROR = "org.jboss.seam.selectitems.entityConverter.emAccessError";
	public static final String INVAlID_SELECTION = "org.jboss.seam.selectitems.entityConverter.invalidSelection";

	/**
	 * This method should not be overridden (non-Javadoc)
	 * 
	 * @see org.jboss.seam.selectitems.jsf.SelectItemsConverter#convertToString(javax.faces.context.FacesContext,
	 *      javax.faces.component.UIComponent, java.lang.Object)
	 */
	@SuppressWarnings("unchecked")
	protected String convertToString(FacesContext faces, UIComponent c,
			Object obj) {
		// For some reason MyFaces will sometimes try to reconvert an already
		// converted objct so
		// pass through
		if (obj instanceof String)
			return (String) obj;
		return getIdAsString((E) obj);
	}

	/**
	 * This method should not be overridden (non-Javadoc)
	 * 
	 * @see org.jboss.seam.selectitems.jsf.SelectItemsConverter#convertToObject(javax.faces.context.FacesContext,
	 *      javax.faces.component.UIComponent, java.lang.String)
	 */
	protected Object convertToObject(FacesContext facesContext, UIComponent c,
			String value) {
		Class clazz = getEntityClass(c, facesContext);
		return loadEntityFromPersistenceContext(clazz, getIdAsObject(value,
				clazz));
	}

	/**
	 * Override this method to provide a custom conversion from Entity to String
	 * It should return a unique, unchanging, reference to the entity
	 * 
	 * @param entity
	 * @return
	 */
	protected abstract String getIdAsString(E entity);

	/**
	 * The class of the entity to load This method can be overriden in a simple
	 * fashion e.g.
	 * 
	 * <code>return MyEntity.class;</code>
	 * 
	 * @param cmp
	 *            The JSF component which the conversion is for
	 * @param facesContext
	 * @return The class of the entity to load,
	 * @throws ConverterException
	 *             If unable to determine the EntityClass
	 */
	protected Class getClazz() {
         Type type = getClass().getGenericSuperclass();
         if (type instanceof ParameterizedType)
         {
            ParameterizedType paramType = (ParameterizedType) type;
            return (Class<E>) paramType.getActualTypeArguments()[0];
         }
         else
         {
            throw new IllegalArgumentException("Could not guess entity class by reflection");
         }
	}
	
	/**
	 * The EntityConverter attempts to guess the entity's class
	 * If the EntityConverter is extended using generics, or is invoked
	 * through si:selectitems or si:convertEntity this should work.
	 * 
	 * Otherwise override getClazz() to specify the class
	 * @param cmp
	 * @param facesContext
	 * @return
	 */
	protected Class getEntityClass(UIComponent cmp, FacesContext facesContext) {
		return getClazz();
	}

	/**
	 * Retrieve the Entity from the PersistenceContext The EnityManagerFactory
	 * specified using persistenceUnitJndiName is used
	 * 
	 * @param clazz
	 *            The class of the entity to load
	 * @param id
	 *            The id of the entity to load
	 * @return The entity
	 * @throws ConverterException
	 *             if retrieval of the entity failed
	 */
	@SuppressWarnings("unchecked")
	private Object loadEntityFromPersistenceContext(Class clazz, Object id) {
		Object entity;
		try {
			EntityManagerFactory emf = (EntityManagerFactory) Naming
					.getInitialContext().lookup(getPersistenceUnitJndiName());
			EntityManager em = emf.createEntityManager();
			entity = em.find(clazz, id);
		} catch (Exception e) {
			throw new ConverterException(
					getMessage(EM_ACCESS_ERROR), e);
		}
		if (entity == null) {
			throw new ConverterException(
					getMessage(INVAlID_SELECTION));
		} else {
			return entity;
		}

	}

	/**
	 * This method should be overriden in a custom EntityConverter. Given the id
	 * (as a String) and the class of the entity, it should return the Id of the
	 * object in its native type
	 * 
	 * @param id
	 *            The Id of the entity
	 * @param clazz
	 *            The class of the entity
	 * @return The id of the entity as a correctly typed object
	 */
	protected abstract Object getIdAsObject(String id, Class clazz);

	/*
	 * ------------------------------------------------ Converter attributes
	 * ------------------------------------------------
	 */

	/*
	 * State saving implementation allowing parameters to be passed
	 * (non-Javadoc)
	 * 
	 * @see javax.faces.component.StateHolder#saveState(javax.faces.context.FacesContext)
	 */

	String persistenceUnitJndiName;

	boolean _transient;

	public Object saveState(FacesContext arg0) {
		Object[] values = new Object[1];
		values[0] = persistenceUnitJndiName;
		return values;
	}

	public void restoreState(FacesContext arg0, Object state) {
		Object[] values = (Object[]) state;
		persistenceUnitJndiName = (String) values[0];
	}

	public boolean isTransient() {
		return _transient;
	}

	public void setTransient(boolean _transient) {
		this._transient = _transient;
	}

	public String getPersistenceUnitJndiName() {
		if (persistenceUnitJndiName == null) {
			return SelectItemsConfig.instance().getPersistenceUnitJndiName();
		} else {
			return persistenceUnitJndiName;
		}
	}

	public void setPersistenceUnitJndiName(String emfJndiName) {
		this.persistenceUnitJndiName = emfJndiName;
	}

}